import numba
import numpy as np

from . import ABS, MAX, REF
from .. import bn_wraps as bn, numba_cache
from .. import talib as ta
from ..nb import numpy_rolling_apply, _rolling_func_1_nb
from ..utils import pd_to_np

_ta1d = ta.init(mode=1, skipna=False, to_globals=False)
_ta2d = ta.init(mode=2, skipna=False, to_globals=False)


def CONST(real):
    """取A最后的值为常量"""
    return np.full_like(real, real[-1])


def SUMIF(real, condition, timeperiod):
    """!!!注意，condition位置"""
    return bn.move_sum(real * condition, window=timeperiod, axis=0)


def TR(high, low, close):
    """TR真实波幅"""
    lc = REF(close, 1)
    return MAX(high - low, ABS(high - lc), ABS(lc - low))


def FILTER(S, N):
    """FILTER函数，S满足条件后，将其后N周期内的数据置为0"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _filter_nb(arr, n):
        is_1d = arr.ndim == 1
        x = arr.shape[0]
        y = 1 if is_1d else arr.shape[1]

        for j in range(y):
            a = arr if is_1d else arr[:, j]

            # 为了跳过不必要部分，由for改while
            i = 0
            while i < x:
                if a[i]:
                    a[i + 1:i + 1 + n] = 0
                    i += n + 1
                else:
                    i += 1
        return arr

    S = pd_to_np(S, copy=True)
    return _filter_nb(S, N)


def BARSLAST(S):
    """BARSLAST(X)，上一次X不为0到现在的天数

    成立当天输出0
    """

    @numba.jit(nopython=True, cache=True, nogil=True)
    def _bars_last_nb(arr, out):
        """上一次条件成立到当前的周期数"""
        is_1d = arr.ndim == 1
        x = arr.shape[0]
        y = 1 if is_1d else arr.shape[1]

        for j in range(y):
            a = arr if is_1d else arr[:, j]
            b = out if is_1d else out[:, j]
            s = 0
            for i in range(x):
                if a[i]:
                    s = 0
                b[i] = s
                s += 1

        return out

    S = pd_to_np(S, copy=False)
    out = np.zeros_like(S, dtype=int)
    return _bars_last_nb(S, out)


def BARSLASTCOUNT(S):
    """BARSLASTCOUNT(X)，统计连续满足X条件的周期数

    成立第一天输出1
    """

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _bars_last_count_nb(arr, out):
        """

        Parameters
        ----------
        arr
        out

        References
        ----------
        https://stackoverflow.com/questions/18196811/cumsum-reset-at-nan

        """
        is_1d = arr.ndim == 1
        x = arr.shape[0]
        y = 1 if is_1d else arr.shape[1]

        for j in range(y):
            a = arr if is_1d else arr[:, j]
            b = out if is_1d else out[:, j]
            s = 0
            for i in range(x):
                if a[i]:
                    s += 1
                    b[i] = s
                else:
                    s = 0

        return out

    S = pd_to_np(S, copy=False)
    out = np.zeros_like(S, dtype=int)
    return _bars_last_count_nb(S, out)


def BARSSINCEN(cond, timeperiod):
    """BARSSINCEN(X,N):N周期内第一次X不为0到现在的天数"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _bars_since_n_nb(a, n):
        """BARSSINCEN(X,N):N周期内第一次X不为0到现在的天数"""
        for i, x in enumerate(a):
            if x:
                return n - i - 1
        return 0

    return numpy_rolling_apply([pd_to_np(cond)], timeperiod, _rolling_func_1_nb, _bars_since_n_nb, timeperiod)



import numba
import numpy as np

from .utils import pd_to_np


def fill_notna(arr, fill_value, n=1):
    """查找第一个非nan,将其改成fill_value"""

    @numba.jit(nopython=True, cache=True, nogil=True)
    def _fill_notna_nb(arr, fill_value, n: int):
        is_1d = arr.ndim == 1
        x = arr.shape[0]
        y = 1 if is_1d else arr.shape[1]

        for j in range(y):
            a = arr if is_1d else arr[:, j]
            k = n
            for i in range(x):
                cur = a[i]
                if cur != cur:
                    continue
                a[i] = fill_value
                k -= 1
                if k <= 0:
                    break

        return arr

    if n < 1:
        return arr
    arr = pd_to_np(arr, copy=True)
    return _fill_notna_nb(arr, fill_value, n)


@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_3_nb(arr0, arr1, arr2, out, timeperiod, func, *args):
    """滚动函数，在二维或三维数上遍历，三组输入

    Parameters
    ----------
    arr0:
        输入。二维或三维
    out:
        输出。一维或二维。降了一个维度
    window: int
        窗口长度
    func:
        单向量处理函数
    args:
        func的位置参数

    Returns
    -------
    np.ndarray
        一维或二维数组

    """
    if arr0.ndim == 3:
        for i, (aa, bb, cc) in enumerate(zip(arr0, arr1, arr2)):
            for j, (a, b, c) in enumerate(zip(aa, bb, cc)):
                out[i + timeperiod - 1, j] = func(a, b, c, *args)
    elif arr0.ndim == 2:
        for i, (a, b, c) in enumerate(zip(arr0, arr1, arr2)):
            out[i + timeperiod - 1] = func(a, b, c, *args)

    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_2_nb(arr0, arr1, out, timeperiod, func, *args):
    """滚动函数，在二维或三维数上遍历，二组输入

    Parameters
    ----------
    arr0:
        输入。二维或三维
    out:
        输出。一维或二维。降了一个维度
    window: int
        窗口长度
    func:
        单向量处理函数
    args:
        func的位置参数

    Returns
    -------
    np.ndarray
        一维或二维数组

    """
    if arr0.ndim == 3:
        for i, (aa, bb) in enumerate(zip(arr0, arr1)):
            for j, (a, b) in enumerate(zip(aa, bb)):
                out[i + timeperiod - 1, j] = func(a, b, *args)
    elif arr0.ndim == 2:
        for i, (a, b) in enumerate(zip(arr0, arr1)):
            out[i + timeperiod - 1] = func(a, b, *args)

    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_1_nb(arr, out, timeperiod, func, *args):
    """滚动函数，在二维或三维数上遍历，一组输入

    Parameters
    ----------
    arr:
        输入。二维或三维
    out:
        输出。一维或二维。降了一个维度
    window: int
        窗口长度
    func:
        单向量处理函数
    args:
        func的位置参数

    Returns
    -------
    np.ndarray
        一维或二维数组

    """
    if arr.ndim == 3:
        for i, aa in enumerate(arr):
            for j, a in enumerate(aa):
                out[i + timeperiod - 1, j] = func(a, *args)
    elif arr.ndim == 2:
        for i, a in enumerate(arr):
            out[i + timeperiod - 1] = func(a, *args)

    return out


def numpy_rolling_apply(inputs, window, func1, func2, *args):
    """滚动应用方法 处理两个"""
    # 输出只与第一个的形状相同
    out = np.empty_like(inputs[0])
    try:
        # 可能出现类似int无法设置nan的情况
        out[:window] = np.nan
    except:
        out[:window] = 0

    arrs = [np.lib.stride_tricks.sliding_window_view(i, window, axis=0) for i in inputs]

    return func1(*arrs, out, window, func2, *args)



"""
公式转alpha101

101 Formulaic Alphas
文档质量一般
1. 部分函数即有全小写，又有大小写混合
2. 很多参数应当是整数，但输入是小数，不得不做修正
"""

from ..imports import wide as W
from ..utils import round_a_i, round_a_a_i

correlation = round_a_a_i(W.CORREL)
decay_linear = round_a_i(W.ts_decay_linear)

LessThan = W.less
rank = W.rank
ts_rank = round_a_i(W.ts_rank)
scale = W.scale
SignedPower = W.signed_power

IF = W.if_else
abs = W.abs_
log = W.log  # 这里是用的自然对数
MAX = W.max_
MIN = W.min_
sign = W.sign

delta = round_a_i(W.ts_delta)
ts_max = round_a_i(W.ts_max)
ts_argmax = round_a_i(W.ts_arg_max)
ts_min = round_a_i(W.ts_min)
ts_argmin = round_a_i(W.ts_arg_min)
product = round_a_i(W.ts_product)
delay = round_a_i(W.ts_delay)
sum = round_a_i(W.ts_sum)

covariance = round_a_a_i(W.ts_covariance)
stddev = round_a_i(W.ts_std_dev)  # 引入的是全体标准差

indneutralize = W.group_neutralize

# 部分别名，这样官方公式可以减少改动
Ts_Rank = ts_rank
IndNeutralize = indneutralize
Ts_ArgMax = ts_argmax
Ts_ArgMin = ts_argmin
LessThan = LessThan
min = MIN
max = MAX
Sign = sign
Log = log



"""
对指标的算子化包装
1. 包装成只支持 长表 输入和输出
2. 简化参数输入，命名参数也可当成位置参数输入
3. 通过dropna的方法，自动跳过停牌

!!!函数太多，又想要智能提示，只能手工按需补充
"""
from .. import talib as ta, BY_ASSET, BY_DATE, BY_GROUP
from ..aggregate import A_div_AB
from ..alphas.alpha import CUMPROD
from ..alphas.alpha import FILTER_191
from ..ema import SMA_CN
from ..regress import REGRESI
from ..regress import SLOPE_YX
from ..tdx.reference import FILTER as FILTER_TDX
from ..tdx.reference import SUMIF
from ..utils import to_pd
from ..utils_long import dataframe_groupby_apply, series_groupby_apply
from ..wq.arithmetic import abs_
from ..wq.arithmetic import log
from ..wq.arithmetic import max_
from ..wq.arithmetic import min_
from ..wq.arithmetic import sign
from ..wq.arithmetic import signed_power
from ..wq.cross_sectional import rank
from ..wq.cross_sectional import scale
from ..wq.group import group_neutralize
from ..wq.logical import if_else
from ..wq.logical import less
from ..wq.time_series import ts_arg_max
from ..wq.time_series import ts_arg_min
from ..wq.time_series import ts_corr
from ..wq.time_series import ts_count
from ..wq.time_series import ts_covariance
from ..wq.time_series import ts_decay_linear
from ..wq.time_series import ts_delay
from ..wq.time_series import ts_delta
from ..wq.time_series import ts_max
from ..wq.time_series import ts_mean
from ..wq.time_series import ts_min
from ..wq.time_series import ts_product
from ..wq.time_series import ts_rank
from ..wq.time_series import ts_std_dev
from ..wq.time_series import ts_sum

# 一维TALIB
_ta1d = ta.init(mode=1, skipna=False, to_globals=False)


## TALIB, 多输入
CORREL = dataframe_groupby_apply(_ta1d.CORREL, by=BY_ASSET)
ATR = dataframe_groupby_apply(_ta1d.ATR, by=BY_ASSET, to_df=[0, 1, 2], to_kwargs={3: 'timeperiod'})

## TALIB, 单输入
LINEARREG_SLOPE = series_groupby_apply(_ta1d.LINEARREG_SLOPE, by=BY_ASSET)
SMA_TA = series_groupby_apply(_ta1d.SMA, by=BY_ASSET)
WMA = series_groupby_apply(_ta1d.WMA, by=BY_ASSET)

# WorldQuant，时序
ts_arg_max = series_groupby_apply(ts_arg_max, by=BY_ASSET, to_kwargs={1: 'd'})
ts_arg_min = series_groupby_apply(ts_arg_min, by=BY_ASSET, to_kwargs={1: 'd'})
ts_corr = dataframe_groupby_apply(ts_corr, by=BY_ASSET, to_kwargs={2: 'd'})
ts_count = series_groupby_apply(ts_count, by=BY_ASSET, to_kwargs={1: 'd'})
ts_covariance = dataframe_groupby_apply(ts_covariance, by=BY_ASSET, to_kwargs={2: 'd'})
ts_decay_linear = series_groupby_apply(ts_decay_linear, by=BY_ASSET, to_kwargs={1: 'd'})
ts_delay = series_groupby_apply(ts_delay, by=BY_ASSET, to_kwargs={1: 'd'})
ts_delta = series_groupby_apply(ts_delta, by=BY_ASSET, to_kwargs={1: 'd'})
ts_max = series_groupby_apply(ts_max, by=BY_ASSET, to_kwargs={1: 'd'})
ts_mean = series_groupby_apply(ts_mean, by=BY_ASSET, to_kwargs={1: 'd'})
ts_min = series_groupby_apply(ts_min, by=BY_ASSET, to_kwargs={1: 'd'})
ts_product = series_groupby_apply(ts_product, by=BY_ASSET, to_kwargs={1: 'd'}, dropna=False)
ts_rank = series_groupby_apply(ts_rank, by=BY_ASSET, to_kwargs={1: 'd'})
ts_std_dev = series_groupby_apply(ts_std_dev, by=BY_ASSET, to_kwargs={1: 'd'})
ts_sum = series_groupby_apply(ts_sum, by=BY_ASSET, to_kwargs={1: 'd'})

# WorldQuant，横截面
rank = series_groupby_apply(rank, by=BY_DATE, to_kwargs={})
scale = series_groupby_apply(scale, by=BY_DATE, to_kwargs={1: 'scale'}, dropna=False)

# WorldQuant
less = to_pd(less)
signed_power = to_pd(signed_power)
if_else = to_pd(if_else)
abs_ = to_pd(abs_)
log = to_pd(log)
max_ = to_pd(max_)
min_ = to_pd(min_)
sign = to_pd(sign)

# 特殊
CUMPROD = series_groupby_apply(CUMPROD, by=BY_ASSET, to_kwargs={})
FILTER_191 = dataframe_groupby_apply(FILTER_191, by=BY_ASSET, to_kwargs={}, dropna=False)
FILTER_TDX = series_groupby_apply(FILTER_TDX, by=BY_ASSET, to_kwargs={1: 'N'})
#


#
SMA_CN = series_groupby_apply(SMA_CN, by=BY_ASSET, to_kwargs={1: 'timeperiod', 2: 'M'})

#

SUMIF = dataframe_groupby_apply(SUMIF, by=BY_ASSET)

SLOPE_YX = dataframe_groupby_apply(SLOPE_YX, by=BY_ASSET)
REGRESI4 = dataframe_groupby_apply(REGRESI, by=BY_ASSET, to_df=[0, 1, 2, 3], to_kwargs={4: 'timeperiod'})

# 行业中性
group_neutralize = group_neutralize

# 可用于 全部市场宽度
A_div_AB_1 = series_groupby_apply(A_div_AB, by=BY_DATE, to_kwargs={})
# 可用于 板块市场宽度
A_div_AB_2 = dataframe_groupby_apply(A_div_AB, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})



"""
Vector Operators

# TODO: 没有搞明白，等看懂了再来实现
"""


def vec_avg(x):
    """Taking mean of the vector field x"""
    pass


def vec_choose(x, nth=1):
    """Choosing kth item(indexed at 0) from each vector field x"""
    pass


def vec_count(x):
    """Number of elements in vector field x"""
    pass


def vec_ir(x):
    """Information Ratio (Mean / Standard Deviation) of vector field x"""
    pass


def vec_kurtosis(x):
    """Kurtosis of vector field x"""
    pass


def vec_max(x):
    """Maximum value form vector field x"""
    pass


def vec_min(x):
    """Minimum value form vector field x"""
    pass


def vec_norm(x):
    """Sum of all absolute values of vector field x"""
    pass


def vec_percentage(x, percentage=0.5):
    """Percentile of vector field x"""
    pass


def vec_powersum(x, constant=2):
    """Sum of power of vector field x"""
    pass


def vec_range(x):
    """Difference between maximum and minimum element in vector field x"""
    pass


def vec_skewness(x):
    """Skewness of vector field x"""
    pass


def vec_stddev(x):
    """Standard Deviation of vector field x"""
    pass


def vec_sum(x):
    """Sum of vector field x"""
    pass



# 此处引入库要小心，有可能打破只想用talib二维库的需求

# ReferenceError: underlying object has vanished
# 如果出现上面错误，请改成False,并删除多个目录下的__pycache__
numba_cache = False

# 按股票分组，计算时序指标。注意，组内时序需要已经排序
BY_ASSET = 'asset'
# 按时间分组。计算横截面
BY_DATE = 'date'
# 横截面上进行行业中性化
BY_GROUP = ['date', 'group']



"""
Transformational Operators
"""
import numba
import numpy as np

from .. import numba_cache
from ..utils import pd_to_np
from ..wq.logical import if_else
from ..wq.time_series import days_from_last_change


def arc_cos(x):
    """If -1 <= x <= 1: arccos(x); else NaN"""
    return np.arccos(x)


def arc_sin(x):
    """If -1 <= x <= 1: arcsin(x); else NaN"""
    return np.arcsin(x)


def arc_tan(x):
    """This operator does inverse tangent of input. """
    return np.arctan(x)


def bucket(x,
           range="0, 1, 0.1",
           buckets="2,5,6,7,10",
           skipBegin=False, skipEnd=False, skipBoth=False,
           NANGroup=True):
    """Convert float values into indexes for user-specified buckets. Bucket is useful for creating group values, which can be passed to group operators as input."""
    pass


def clamp(x, lower=0, upper=0, inverse=False, mask=np.nan):
    """Limits input value between lower and upper bound in inverse = false mode (which is default). Alternatively, when inverse = true, values between bounds are replaced with mask, while values outside bounds are left as is."""
    if inverse:
        # mask is one of: 'nearest_bound', 'mean', 'NAN' or any floating point number
        return if_else((x > lower) & (x < upper), mask, x)
    else:
        # q = if_else(x < lower, lower, x)
        # u = if_else(q > upper, upper, q)
        return np.clip(x, lower, upper)


def filter(x, h="1, 2, 3, 4", t="0.5"):
    """Used to filter the value and allows to create filters like linear or exponential decay."""
    pass


def keep(x, f, period=5):
    """This operator outputs value x when f changes and continues to do that for “period” days after f stopped changing. After “period” days since last change of f, NaN is output."""
    D = days_from_last_change(f)
    return trade_when(D < period, x, D > period)


def left_tail(x, maximum=0):
    """NaN everything greater than maximum, maximum should be constant."""
    return np.where(x > maximum, np.nan, x)


def pasteurize(x):
    """Set to NaN if x is INF or if the underlying instrument is not in the Alpha universe"""
    # TODO: 不在票池中的的功能无法表示
    x = x.copy()
    x[np.isinf(x)] = np.nan
    return x


def right_tail(x, minimum=0):
    """NaN everything less than minimum, minimum should be constant."""
    return np.where(x < minimum, np.nan, x)


def sigmoid(x):
    """Returns 1 / (1 + exp(-x))"""
    return 1 / (1 + np.exp(-x))


def tail(x, lower=0, upper=0, newval=0):
    """If (x > lower AND x < upper) return newval, else return x. Lower, upper, newval should be constants. """
    return np.where((x > lower) & (x < upper), newval, x)


def tanh(x):
    """Hyperbolic tangent of x"""
    return np.tanh(x)


def trade_when(x, y, z):
    """Used in order to change Alpha values only under a specified condition and to hold Alpha values in other cases. It also allows to close Alpha positions (assign NaN values) under a specified condition."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _trade_when_nb(xx, yy, zz, out):
        is_1d = xx.ndim == 1
        x = xx.shape[0]
        y = 1 if is_1d else xx.shape[1]

        for j in range(y):
            a = xx if is_1d else xx[:, j]
            b = yy if is_1d else yy[:, j]
            c = zz if is_1d else zz[:, j]
            d = out if is_1d else out[:, j]
            last = np.nan
            for i in range(x):
                if c[i] > 0:
                    d[i] = np.nan
                elif a[i] > 0:
                    d[i] = b[i]
                else:
                    d[i] = last
                last = d[i]

        return out

    x = pd_to_np(x, copy=False)
    y = pd_to_np(y, copy=False)
    z = pd_to_np(z, copy=False)
    out = np.empty_like(y)
    return _trade_when_nb(x, y, z, out)



"""
Special Operators
"""


def convert(x, mode="dollar2share"):
    """Convert dollars to share or share to dollar when mode = "share2dollar"
    """
    pass


def inst_pnl(x):
    """Generate pnl per instruments"""
    pass



"""
Arithmetic Operators
"""
from functools import reduce

import numpy as np

from .time_series import ts_delta


def abs_(x):
    """Absolute value of x"""
    return np.absolute(x)


def add(*args, filter=False):
    """Add all inputs (at least 2 inputs required). If filter = true, filter all input NaN to 0 before adding"""

    def _add(x, y):
        mask = np.isnan(x) & np.isnan(y)
        x = np.nan_to_num(x, nan=0)
        y = np.nan_to_num(y, nan=0)
        x = np.ma.masked_array(x, mask=mask)
        y = np.ma.masked_array(y, mask=x.mask)
        return (x + y).filled(np.nan)

    func = _add if filter else np.add

    with np.errstate(over='ignore'):
        return reduce(func, args)


def ceiling(x):
    """Nearest larger integer"""
    return np.ceil(x)


def divide(*args):
    """x / y

    RuntimeWarning: invalid value encountered in true_divide
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        return reduce(np.true_divide, args)


def exp(x):
    """Natural exponential function: e^x"""
    return np.exp(x)


def floor(x):
    """Nearest smaller integer"""
    return np.floor(x)


def fraction(x):
    """This operator removes the whole number part and returns the remaining fraction part with sign."""
    # return sign(x) * (abs(x) - floor(abs(x)))
    return sign(x) * (abs_(x) % 1.)


def inverse(x):
    """1 / x"""
    x = x.copy()
    x[x == 0] = np.nan
    return 1 / x


def log(x):
    """Natural logarithm. For example: Log(high/low) uses natural logarithm of high/low ratio as stock weights."""
    return np.log(np.where(x > 0, x, np.nan))


def log_diff(x):
    """Returns log(current value of input or x[t] ) - log(previous value of input or x[t-1])."""
    return ts_delta(log(x), 1)


def max_(*args):
    """Maximum value of all inputs. At least 2 inputs are required."""
    return reduce(np.maximum, args)


def min_(*args):
    """Minimum value of all inputs. At least 2 inputs are required."""
    return reduce(np.minimum, args)


def multiply(*args, filter=False):
    """Multiply all inputs. At least 2 inputs are required. Filter sets the NaN values to 1."""

    def _multiply(x, y):
        mask = np.isnan(x) & np.isnan(y)
        x = np.nan_to_num(x, nan=1)
        y = np.nan_to_num(y, nan=1)
        x = np.ma.masked_array(x, mask=mask)
        y = np.ma.masked_array(y, mask=x.mask)
        return (x * y).filled(np.nan)

    func = _multiply if filter else np.multiply

    with np.errstate(over='ignore'):
        return reduce(func, args)


def nan_mask(x, y):
    """replace input with NAN if input's corresponding mask value or the second input here, is negative."""
    x = x.copy()
    x[y < 0] = np.nan
    return x


def nan_out(x, lower=0, upper=0):
    """If x < lower or x > upper return NaN, else return x. At least one of "lower", "upper" is required."""
    return np.where((x < lower) | (x > upper), np.nan, x)


def power(x, y):
    """x ^ y"""
    with np.errstate(divide='ignore', invalid='ignore'):
        r = np.power(x, y)
        # 有可能产生inf，是否需要处理
        # r[np.isinf(r)] = np.nan
        return r


def purify(x):
    """Clear infinities (+inf, -inf) by replacing with NaN."""
    x = x.copy()
    x[np.isinf(x)] = np.nan
    return x


def replace(x, target=[-np.inf, np.inf], dest=[np.nan, np.nan]):
    """Replace target values in input with destination values."""
    x = x.copy()
    for t, d in zip(target, dest):
        x[x == t] = d
    return x


def reverse(x):
    """- x"""
    return -x


def round_(x):
    """Round input to closest integer."""
    return np.around(x)


def round_down(x, f=1):
    """Round input to greatest multiple of f less than input;"""
    with np.errstate(invalid='ignore', divide='ignore'):
        return x // f * f


def sign(x):
    """if input = NaN; return NaN
else if input > 0, return 1
else if input < 0, return -1
else if input = 0, return 0

    return (0 < x) * 1 - (x < 0)
    """
    return np.sign(x)


def signed_power(x, y):
    """x raised to the power of y such that final result preserves sign of x."""
    with np.errstate(invalid='ignore', divide='ignore'):
        return sign(x) * (abs_(y) ** y)


def s_log_1p(x):
    """Confine function to a shorter range using logarithm such that higher input remains higher and negative input remains negative as an output of resulting function and -1 or 1 is an asymptotic value."""
    with np.errstate(invalid='ignore', divide='ignore'):
        return sign(x) * log(1 + abs_(x))


def sqrt(x):
    """Square root of x"""
    with np.errstate(invalid='ignore'):
        return np.sqrt(x)


def subtract(*args, filter=False):
    """x-y. If filter = true, filter all input NaN to 0 before subtracting"""

    def _subtract(x, y):
        mask = np.isnan(x) & np.isnan(y)
        x = np.nan_to_num(x, nan=0)
        y = np.nan_to_num(y, nan=0)
        x = np.ma.masked_array(x, mask=mask)
        y = np.ma.masked_array(y, mask=x.mask)
        return (x - y).filled(np.nan)

    func = _subtract if filter else np.subtract

    with np.errstate(over='ignore'):
        return reduce(func, args)


def to_nan(x, value=0, reverse=False):
    """Convert value to NaN or NaN to value if reverse=true"""
    x = x.copy()
    if reverse:
        x[np.isnan(x)] = value
    else:
        x[x == value] = np.nan
    return x


def densify(x):
    """Converts a grouping field of many buckets into lesser number of only available buckets so as to make working with grouping fields computationally efficient."""
    return x


# ----------------
def log10(x):
    return np.log10(np.where(x > 0, x, np.nan))


# 过于简单，直接添加于此
def mean(*args, filter=True):
    """多个均值"""
    return add(*args, filter=filter) / len(args)



import numba
import numpy as np

from .. import bn_wraps as bn, numba_cache
from .. import talib as ta
from ..nb import numpy_rolling_apply, _rolling_func_1_nb
from ..utils import pd_to_np

_ta1d = ta.init(mode=1, skipna=False)
_ta2d = ta.init(mode=2, skipna=False)


def AVEDEV(real, timeperiod: int):
    """平均绝对偏差

    AVEDEV(real, timeperiod=20)
    """

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _avedev_nb(a):
        """avedev平均绝对偏差"""
        return np.mean(np.abs(a - np.mean(a)))

    return numpy_rolling_apply([pd_to_np(real)], timeperiod, _rolling_func_1_nb, _avedev_nb)


def STD(x, d):
    """样本标准差"""
    return bn.move_std(x, window=d, axis=0, ddof=1)


def STDP(x, d):
    """总体标准差"""
    return bn.move_std(x, window=d, axis=0, ddof=0)


def VAR(x, d):
    """样本方差"""
    return bn.move_var(x, window=d, axis=0, ddof=1)


def VARP(x, d):
    """总体方差"""
    return bn.move_var(x, window=d, axis=0, ddof=0)



"""
Time Series Operators
"""
import numba
import numpy as np
from numba.typed import List

from .. import bn_wraps as bn, numba_cache
from .. import talib as ta
from ..nb import numpy_rolling_apply, _rolling_func_1_nb, _rolling_func_2_nb, _rolling_func_3_nb
from ..utils import pd_to_np

_ta1d = ta.init(mode=1, skipna=False, to_globals=False)
_ta2d = ta.init(mode=2, skipna=False, to_globals=False)


def days_from_last_change(x):
    """Amount of days since last change of x"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _days_from_last_change_nb(arr, out):
        is_1d = arr.ndim == 1
        x = arr.shape[0]
        y = 1 if is_1d else arr.shape[1]

        for j in range(y):
            a = arr if is_1d else arr[:, j]
            b = out if is_1d else out[:, j]
            last = a[0]
            k = 0
            for i in range(1, x):
                cur = a[i]
                if cur == last:
                    k += 1
                else:
                    last = cur
                    k = 0
                b[i] = k

        return out

    x = pd_to_np(x, copy=False)
    out = np.zeros_like(x, dtype=int)
    return _days_from_last_change_nb(x, out)


def ts_weighted_delay(x, k=0.5):
    """Instead of replacing today’s value with yesterday’s as in ts_delay(x, 1), it assigns weighted average of today’s and yesterday’s values with weight on today’s value being k and yesterday’s being (1-k)."""
    pass


def hump(x, hump=0.01):
    """Limits amount and magnitude of changes in input (thus reducing turnover)."""
    pass


def hump_decay(x, p=0, relative=False):
    """This operator helps to ignore the values that changed too little corresponding to previous ones."""
    t1 = ts_delay(x, 1)
    t2 = abs(x - t1)
    if relative:
        t3 = abs(x + t1)
        return np.where(t2 > (p * t3), x, t1)
    else:
        return np.where(t2 > p, x, t1)

def inst_tvr(x, d):
    """Total trading value / Total holding value in the past d days"""
    pass


def jump_decay(x, d, stddev=False, sensitivity=0.5, force=0.1):
    """If there is a huge jump in current data compare to previous one"""
    std = 1 if not stddev else ts_std_dev(x, d)
    t0 = ts_delay(x, 1)
    t1 = abs(x - t0) > sensitivity * std
    t2 = t0 + (x - t0) * force

    return np.where(t1, t2, x)


def kth_element(x, d, k=1, ignore=[]):
    """Returns K-th value of input by looking through lookback days. This operator can be used to backfill missing data if k=1"""

    # 由于原文档中数据需要[::-1]，所以backfill其实是ffill
    # ignore参数用来计数时跳过，文档不对应
    # TODO: 需要再验证
    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _kth_element_nb(arr, k, ignore):
        for a in arr[-1 - k::-1]:
            if a == a:
                for i in ignore:
                    if a == i:
                        continue
                    else:
                        return a
            else:
                for i in ignore:
                    if i != i:
                        continue

    return numpy_rolling_apply([pd_to_np(x)],
                               d, _rolling_func_1_nb, _kth_element_nb,
                               k, List(ignore))


def last_diff_value(x, d):
    """Returns last x value not equal to current x value from last d days"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _last_diff_value_nb(arr):
        l = arr[-1]
        for a in arr[::-1]:
            if a != l:
                return a
        return l

    return numpy_rolling_apply([pd_to_np(x)],
                               d, _rolling_func_1_nb, _last_diff_value_nb)


def ts_arg_max(x, d):
    """Returns the relative index of the min value in the time series for the past d days. If the current day has the min value for the past d days, it returns 0. If previous day has the min value for the past d days, it returns 1."""
    # 文档中的数据其实要[::-1]
    return bn.move_argmax(x, window=d, axis=0)


def ts_arg_min(x, d):
    """Returns the relative index of the min value in the time series for the past d days; If the current day has the min value for the past d days, it returns 0; If previous day has the min value for the past d days, it returns 1"""
    return bn.move_argmin(x, window=d, axis=0)


def ts_av_diff(x, d):
    """Returns x - tsmean(x, d), but deals with NaNs carefully. That is NaNs are ignored during mean computation."""

    def _ts_mean(x, d):
        """Returns average value of x for the past d days."""
        # 这里为了跳过空值
        t1 = bn.move_sum(np.nan_to_num(x), window=d, axis=0)
        t2 = bn.move_sum(~np.isnan(x), window=d, axis=0)
        return t1 / t2

    # 平均数的计算要跳过NaN
    return x - _ts_mean(x, d)


def ts_backfill(x, lookback=10, k=1, ignore="NAN"):
    """Backfill is the process of replacing the NAN or 0 values by a meaningful value (i.e., a first non-NaN value)"""
    pass


def ts_co_kurtosis(y, x, d):
    """Returns cokurtosis of y and x for the past d days."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _co_kurtosis_nb(arr0, arr1):
        t1 = arr0 - np.nanmean(arr0)
        t2 = arr1 - np.nanmean(arr1)
        t3 = np.nanstd(arr0)
        t4 = np.nanstd(arr1)
        return np.nanmean(t1 * (t2 ** 3)) / (t3 * (t4 ** 3))

    return numpy_rolling_apply([pd_to_np(y), pd_to_np(x)],
                               d, _rolling_func_2_nb, _co_kurtosis_nb)


def ts_corr(x, y, d):
    """Returns correlation of x and y for the past d days"""
    if True:
        return _ta2d.CORREL(x, y, timeperiod=d)
    else:
        @numba.jit(nopython=True, cache=numba_cache, nogil=True)
        def _corr_nb(arr0, arr1):
            return np.corrcoef(arr0, arr1)[0, 1]

        return numpy_rolling_apply([pd_to_np(x), pd_to_np(y)],
                                   d, _rolling_func_2_nb, _corr_nb)


def ts_co_skewness(y, x, d):
    """Returns coskewness of y and x for the past d days."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _co_skewness_nb(arr0, arr1):
        t1 = arr0 - np.nanmean(arr0)
        t2 = arr1 - np.nanmean(arr1)
        t3 = np.nanstd(arr0)
        t4 = np.nanstd(arr1)
        return np.nanmean(t1 * (t2 ** 2)) / (t3 * (t4 ** 2))

    return numpy_rolling_apply([pd_to_np(y), pd_to_np(x)],
                               d, _rolling_func_2_nb, _co_skewness_nb)


def ts_count_nans(x, d):
    """Returns the number of NaN values in x for the past d days"""
    return ts_sum(np.isnan(x), d)


def ts_covariance(y, x, d):
    """Returns covariance of y and x for the past d days"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _covariance_nb(arr0, arr1):
        """协方差矩阵"""
        return np.cov(arr0, arr1)[0, 1]

    return numpy_rolling_apply([pd_to_np(y), pd_to_np(x)],
                               d, _rolling_func_2_nb, _covariance_nb)


def ts_decay_exp_window(x, d, factor=0.5):
    """Returns exponential decay of x with smoothing factor for the past d days."""
    pass


def ts_decay_linear(x, d, dense=True):
    """Returns the linear decay on x for the past d days. Dense parameter=false means operator works in sparse mode and we treat NaN as 0. In dense mode we do not."""
    # TODO: 这样写是否合适？
    if dense:
        return _ta2d.WMA(x, timeperiod=d)
    else:
        return _ta2d.WMA(np.nan_to_num(x), timeperiod=d)


def ts_delay(x, d):
    """Returns x value d days ago"""
    if d == 0 or np.isnan(d):
        # 该不该复制呢？
        return x
    arr = np.empty_like(x)
    if d > 0:
        arr[:d] = np.nan
        arr[d:] = x[:-d]
    if d < 0:
        # 为了复刻shift(-n)
        arr[d:] = np.nan
        arr[:d] = x[-d:]
    return arr


def ts_delta(x, d):
    """Returns x - ts_delay(x, d)"""
    return x - ts_delay(x, d)


def ts_ir(x, d):
    """Return information ratio ts_mean(x, d) / ts_std_dev(x, d)"""
    return ts_mean(x, d) / ts_std_dev(x, d)


def ts_kurtosis(x, d):
    """Returns kurtosis of x for the last d days."""

    # # 与scipy.stats.kurtosis结果相同，与pd.rolling.kurt结果不同, bias的问题
    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _kurtosis_nb(arr):
        t1 = arr - np.nanmean(arr)
        return np.nanmean(t1 ** 4) / (np.nanmean(t1 ** 2) ** 2)

    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _kurtosis_nb) - 3


def ts_max(x, d):
    """Returns max value of x for the past d days"""
    return bn.move_max(x, window=d, axis=0)


def ts_max_diff(x, d):
    """Returns x - ts_max(x, d)"""
    return x - ts_max(x, d)


def ts_mean(x, d):
    """Returns average value of x for the past d days."""
    return bn.move_mean(x, window=d, axis=0)


def ts_median(x, d):
    """Returns median value of x for the past d days"""
    return bn.move_median(x, window=d, axis=0)


def ts_min(x, d):
    """Returns min value of x for the past d days"""
    return bn.move_min(x, window=d, axis=0)


def ts_min_diff(x, d):
    """Returns x - ts_min(x, d)"""
    return x - ts_min(x, d)


def ts_min_max_cps(x, d, f=2):
    """Returns (ts_min(x, d) + ts_max(x, d)) - f * x. If not specified, by default f = 2"""
    return (ts_min(x, d) + ts_max(x, d)) - f * x


def ts_min_max_diff(x, d, f=0.5):
    """Returns x - f * (ts_min(x, d) + ts_max(x, d)). If not specified, by default f = 0.5"""
    return x - f * (ts_min(x, d) + ts_max(x, d))


def ts_moment(x, d, k=0):
    """Returns K-th central moment of x for the past d days."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _moment_nb(arr, k):
        """中心矩阵"""
        return np.nanmean((arr - np.nanmean(arr)) ** k)

    if k == 2:
        # 结果是一样的，使用计算更快的版本
        return bn.move_var(x, window=d, axis=0)
    # 计算K阶中心矩
    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _moment_nb, k)


def ts_partial_corr(x, y, z, d):
    """Returns partial correlation of x, y, z for the past d days."""

    # TODO: 不知道是否写正确，需要check
    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _partial_corr_nb(arr0, arr1, arr2):
        c = np.corrcoef(np.vstack((arr0, arr1, arr2)))
        rxy = c[0, 1]
        rxz = c[0, 2]
        ryz = c[1, 2]
        t1 = rxy - rxz * ryz
        t2 = (1 - rxz ** 2) ** 0.5
        t3 = (1 - ryz ** 2) ** 0.5
        return t1 / (t2 * t3)

    return numpy_rolling_apply([pd_to_np(x), pd_to_np(y), pd_to_np(z)],
                               d, _rolling_func_3_nb, _partial_corr_nb)


def ts_percentage(x, d, percentage=0.5):
    """Returns percentile value of x for the past d days."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _percentage_nb(a, percentage):
        return np.quantile(a, percentage)

    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _percentage_nb, percentage)


def ts_poly_regression(y, x, d, k=1):
    """Returns y - Ey, where Ey = x + x^2 + … + x^k over d days; k must be specified as a keyword argument"""
    pass


def ts_product(x, d):
    """Returns product of x for the past d days"""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _product_nb(a):
        return np.nanprod(a)

    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _product_nb)


def ts_rank(x, d, constant=0):
    """Rank the values of x for each instrument over the past d days, then return the rank of the current value + constant. If not specified, by default, constant = 0."""
    t1 = bn.move_rank(x, window=d, axis=0)
    return (t1 + 1) / 2 + constant


def ts_regression(y, x, d, lag=0, rettype=0):
    """Returns various parameters related to regression function."""
    pass


def ts_returns(x, d, mode=1):
    """Returns the relative change in the x value ."""
    t1 = ts_delay(x, d)
    if mode == 1:
        return (x - t1) / t1
    if mode == 2:
        return (x - t1) / ((x + t1) / 2)


def ts_scale(x, d, constant=0):
    """Returns (x – ts_min(x, d)) / (ts_max(x, d) – ts_min(x, d)) + constant
This operator is similar to scale down operator but acts in time series space."""
    t1 = ts_min(x, d)
    t2 = ts_max(x, d)
    return (x - t1) / (t2 - t1) + constant


def ts_skewness(x, d):
    """Return skewness of x for the past d days."""

    # 与scipy.stats.skew结果相同，与pd.rolling.skew结果不同, bias的问题
    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _skewness_nb(arr):
        t1 = arr - np.nanmean(arr)
        return np.nanmean(t1 ** 3) / (np.nanmean(t1 ** 2) ** 1.5)

    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _skewness_nb)


def ts_std_dev(x, d):
    """Returns standard deviation of x for the past d days"""
    # 长表与宽表计算居然结果有微小差异
    # return bn.move_std(x, window=d, axis=0, ddof=0)
    if x.ndim == 2:
        return _ta2d.STDDEV(x, timeperiod=d)
    else:
        return _ta1d.STDDEV(x, timeperiod=d)


def ts_step(d):
    """Returns days' counter"""
    # TODO: 这种如何应用是个问题
    return np.arange(d)


def ts_sum(x, d):
    """Sum values of x for the past d days."""
    if d > 0:
        return bn.move_sum(x, window=d, axis=0)
    else:
        # 0表示全量累计求和，这是为通达信公式服务
        return np.cumsum(x, axis=0)


def ts_theilsen(x, y, d):
    """Theil Sen slope estimator of inputs for the past n days."""
    pass


def ts_triple_corr(x, y, z, d):
    """Returns triple correlation of x, y, z for the past d days."""

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _triple_corr_nb(arr0, arr1, arr2):
        t1 = arr0 - np.nanmean(arr0)
        t2 = arr1 - np.nanmean(arr1)
        t3 = arr2 - np.nanmean(arr2)
        t4 = np.nanstd(arr0)
        t5 = np.nanstd(arr1)
        t6 = np.nanstd(arr2)
        return np.nanmean(t1 * t2 * t3) / (t4 * t5 * t6)

    return numpy_rolling_apply([pd_to_np(x), pd_to_np(y), pd_to_np(z)],
                               d, _rolling_func_3_nb, _triple_corr_nb)


def ts_zscore(x, d):
    """Z-score is a numerical measurement that describes a value's relationship to the mean of a group of values. Z-score is measured in terms of standard deviations from the mean: (x - tsmean(x,d)) / tsstddev(x,d)"""
    return (x - ts_mean(x, d)) / ts_std_dev(x, d)


def ts_entropy(x, d, buckets=10):
    """For each instrument, we collect values of input in the past d days and calculate the probability distribution then the information entropy via a histogram as a result."""
    pass


def ts_vector_neut(x, y, d):
    """Returns x- ts_vector_proj(x,y,d)"""
    return x - ts_vector_proj(x, y, d)


def ts_vector_proj(x, y, d):
    """Returns vector projection of x onto y in time-series space. Algebraic and geometric details can be found on wiki"""
    pass


def ts_rank_gmean_amean_diff(input1, input2, input3, d):
    """Returns Geometric Mean of ts_rank(input,d) of all input - Arithmetic Mean of ts_rank(input,d) of all input. This is similar to rank_gmean_amean_diff operator but in time-series space."""
    pass


def ts_quantile(x, d, driver="gaussian"):
    """It calculates ts_rank and apply to its value an inverse cumulative density function from driver distribution. Possible values of driver (optional ) are "gaussian", "uniform", "cauchy" distribution where "gaussian" is the default."""

    # 与ts_percentage区别是什么，缺失一个参数是什么？
    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _quantile_nb(a, q):
        return np.quantile(a, q)

    return numpy_rolling_apply([pd_to_np(x)], d, _rolling_func_1_nb, _quantile_nb, 0.5)


# ------------------------


# 其它功能别名
ts_count = ts_sum



"""
# 计算说明
计算是需要在 `快` 与 `精` 中进行平衡，大规模计算时使用`快速评估`，然后再使用工具进行高精度回测

## 核心
1. 对于单一投资品的收益率，对数收益率时序上可加
2. 对于不同投资品的收益率，应该用简单收益率，因为它在截面上有可加性

## fast，快速计算
1. 不考虑手续费、滑点等信息，否则代码过于复杂。一般用于大规模挖掘时使用，挖出的结果再进行中高精度计算
2. 使用对数收益率累加比简单收益率累乘，计算更快
3. 信号操作需要转换成持仓状态才能使用。信号一般要下移一格才是持仓
4. 做空时收益计算有误差，但在涨跌幅不大时，可以忽略
5. 收益计算，本质上是段内单利，段间复利。但以前使用的收益率累乘，相当于每天调仓为一段，变成了每天复利，
    所以如果代码是每月调仓，收益率就得换成每月，如果是不规则调仓，那就使用 (出场价/入场价-1) 进行计算
6. 持仓取值是1 0 -1还是0.5一类的呢？本质上是看操作的资金占总资金的比重
7. 如果持仓是-2 0 1 3等手数，那盈亏其实应当用diff差分，但这种情况下建议使用vectorbt

## medium, 中速计算
1. 使用快速的向量计算框架vectorbt
2. 可以灵活设置交易成本等信息

## slow, 慢速框架
1. 通常使用事件框架来实现。如backtrader、rqalpha、zipline、聚宽等
2. 与这些平台的交互都统一成因子值，此处不存因子表达式
3. 这些平台都当成算法交易执行工具进行回测

"""
import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy import stats

from . import bn_wraps as bn
from .utils import np_to_pd


def to_log_returns(prices: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """对数收益率

    Parameters
    ----------
    prices:
        价格

    Returns
    -------
    对数收益率，时序上可加

    See Also
    --------
    ffn

    """
    prices = np_to_pd(prices).replace(0, np.nan)
    return np.log(prices / prices.shift(1))


def to_simple_returns(prices: Union[pd.Series, pd.DataFrame], sub_one=True) -> Union[pd.Series, pd.DataFrame]:
    """简单收益率

    Parameters
    ----------
    prices
    sub_one
        是否要在内部减1。减1是收益率，不减是价格比

    Returns
    -------
    简单收益率，横截面上可加

    See Also
    --------
    ffn

    """
    prices = np_to_pd(prices).replace(0, np.nan)
    if sub_one:
        return prices / prices.shift(1) - 1.
    else:
        return prices / prices.shift(1)


def log_returns_cumsum(returns: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """对数收益率累加

    Parameters
    ----------
    returns:
        简单收益率 / 对数收益率

    Notes
    -----
    - 简单收益率: cumsum得到的结果是单利。不建议用于做空
    - 对数收益率: cumsum得到的是log(pn/p1)，本质上是复利

    """
    returns = np_to_pd(returns)
    return returns.fillna(0.0).cumsum()


def simple_returns_cumprod(returns: Union[pd.Series, pd.DataFrame], add_one=True) -> Union[pd.Series, pd.DataFrame]:
    """简单收益率累乘

    Parameters
    ----------
    returns
        简单收益率或价格比
    add_one
        是否内存加1。输入是收益率，需要加1，输入是价格比，可省去加1

    Returns
    -------
    还原价格

    Notes
    -----
    - 简单收益率：cumprod得到的是复利
    - 对数收益率：不能累乘

    """
    returns = np_to_pd(returns)
    if add_one:
        # 简单收益率(减一后的值)
        return returns.add(1., fill_value=0.0).cumprod()
    else:
        # 价格比(nan需要改成1) 1表示P1/P1
        return returns.fillna(1.).cumprod()


def log_returns_exp(returns: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """对数收益率还原。还原后是Pn/P1比值，相当于净值

    to_returns(df)+1 = returns_exp(to_log_returns(df))

    -1 = log(exp(-1))
    1 = exp(log(1))

    log定义域>0

    Parameters
    ----------
    returns
        - 输入简单收益率：原数据没有经过log处理，还原无意义
        - 输入对数收益率：还原的只是价格比，-1后才是简单收益率
        - 累加对数收益率。还原后是以1为起点的净值

    """
    return np.exp(returns)


def equal_weighted_index(prices=None, returns=None, need_one=False) -> pd.Series:
    """等权重指数

    将多条价格或净值等权重合成一条指数

    1. 价格得到简单收益率。只有简单收益率才能横截面加和
    2. 得到横截面平均。自动跳过NaN
    3. 简单收益率只能累乘求复利
    4. 没有做空，一直是buy and hold，可以使用 简单收益率累乘复利

    Parameters
    ----------
    prices
        价格或净值
    returns
        收益率或价格比
    need_one
        收益率是否要加1。如果已经是比值了就不需


    Returns
    -------
    pd.Series
        指数或合成净值

    """
    if returns is None:
        need_one = False  # 可以加快计算
        returns = to_simple_returns(prices, need_one)
    # 不进行加1减1的操作，加快速度
    # 虽然是累乘，但由于只计算一列，所以速度还能接受
    return simple_returns_cumprod(returns.mean(axis=1), need_one)


def weighted_index(weights, prices=None, returns=None, need_one=False):
    """加权指数

    Parameters
    ----------
    prices
        价格
    weights
        权重
    returns
        收益率或价格比
    need_one
        收益率是否要加1。如果已经是比值了就不需


    Returns
    -------
    加权指数

    """
    if returns is None:
        need_one = False  # 可以加快计算
        returns = to_simple_returns(prices, need_one)
    else:
        returns = np_to_pd(returns)

    if need_one:
        returns = returns.fillna(0)
    else:
        returns = returns.fillna(1)

    weights = np_to_pd(weights)
    weights = weights.div(weights.abs().sum(axis=1), axis=0).fillna(0)
    return simple_returns_cumprod((returns * weights).sum(axis=1), need_one)


def ic(factor, returns, rank=True):
    """IC信息系数(Information Coefficient)

    输入一维时，输出标量
    输入二维时，输出一维序列，本质上是横截面上计算IC

    Parameters
    ----------
    factor
    returns
    rank

    Returns
    -------

    """
    x = returns
    y = factor
    axis = x.ndim - 1

    if rank:
        y = bn.nanrankdata(y, axis=axis)
        x = bn.nanrankdata(x, axis=axis)

    if y.ndim == 1:
        return np.corrcoef(y, x)[0, 1]
    else:
        up = bn.nanmean(x * y, axis=axis) - bn.nanmean(x, axis=axis) * bn.nanmean(y, axis=axis)
        down = bn.nanstd(x, axis=axis) * bn.nanstd(y, axis=axis)
        return up / down


def ir(ics):
    """IR即信息比率（Information Ratio）

    IR有两种：

    - 因子IR: IC均值/IC标准差
    - 策略IR: 超额收益/超额收益波动率

    Parameters
    ----------
    ics: array_like
        IC序列

    Returns
    -------
    double
        IR值

    Examples
    --------


    """
    return bn.nanmean(ics) / bn.nanstd(ics)


def ic_decay(factor, returns, rank=True, lag=[1, 2, 3, 4]):
    """IC衰减

    Parameters
    ----------
    factor: pd.DataFrame
        当期因子
    returns: pd.DataFrame
        下期收益。请认真考虑是否要shift
    rank: str
        - spearman: 相当于 ``rank(pct=False)`` ,然后 `pearson`
        - pearson:
    lag: list
        延后期数

        - 0: 表示当期因子，当期收益。使用了未来数据，不合法
        - 1: 表示当期因子，下一期收益

    Returns
    -------
    pd.DataFrame
        IC衰减值

    Examples
    --------
    >>> ics = ic_decay(factor, returns, lag=range(1, 5))
    >>> ics.mean().plot.bar()

    """
    returns = np_to_pd(returns)

    ics = {}
    for i in lag:
        if i == 0:
            warnings.warn('0：表示当期因子，当期收益。使用了未来数据')
        ics[i] = ic(factor, returns.shift(1 - i), rank)
    return pd.DataFrame(ics)


def ttest(ics):
    """t检验。

    Parameters
    ----------
    ics: 1d array
        一维序列

    Returns
    -------

    """
    t_stat, p_value = stats.ttest_1samp(ics, 0, nan_policy='omit')
    return t_stat, p_value


def turnover(weight, periods=1):
    """滚动换手率=(老仓-新仓)/老仓

    Parameters
    ----------
    weight : array_like
        每股仓位
    periods : int, optional
        滚动计算周期，默认值为1，即为计算每条截面数据的滚动换手率

    Returns
    -------
    array_like:
        滚动换手率

    References
    ----------
    https://github.com/quantopian/alphalens/blob/master/alphalens/performance.py#L568

    """
    weight = np_to_pd(weight)
    # nan与0都要标记成0，其它的都要标记成1
    w = (weight.fillna(0) != 0).astype(int)

    """
    a,b,c,d,e
    1,1,0,0,0
    0,1,0,1,1

    -1,0,0,1,1
    
    -1 表示被调出
    1  表示被调入
    0  表示不动，可能在池中，也可能在池外
    """
    # 调入数量/原数量
    return (w.diff(periods) == 1).sum(axis=1) / (w != 0).sum(axis=1)



import numpy as _np


def A_div_AB(x):
    """A/(A+B)

    输入一维，得到一个值
    输入二维，得到一列值

    可用于计算市场宽度等指标
    """
    if x.ndim == 2:
        t1 = _np.nansum(x, axis=1, keepdims=True)
        t2 = _np.nansum(~_np.isnan(x), axis=1, keepdims=True)
    else:
        t1 = _np.nansum(x)
        t2 = _np.nansum(~_np.isnan(x))
    return t1 / t2



import numpy as np

from .wq.cross_sectional import rank as RANK


def cut(x, bins=[0, 10, 100, 1000, 10000], rank=True, pct=False):
    """返回的值最小值一般是1，但如果原值小于bins，那么最小值是0"""
    if rank:
        x = RANK(x, pct=pct)
    return np.digitize(x, bins, right=True)


def qcut(x, q=[0, 0.5, 1], rank=True, pct=True):
    if rank:
        x = RANK(x, pct=pct)
    return np.digitize(x, bins=q, right=True)


def top_k(x, bins=[0, 50, 100, 200]):
    """前N权重字典。正序，越小排名越靠前。所以用户可能需要主动取负数

    Parameters
    ----------
    x
    bins

    """
    labels = cut(x, bins=bins, rank=True, pct=False)
    # 前200=前50+前50到100+前100到200
    d = {k: np.where(labels <= i, x, np.nan) for i, k in enumerate(bins) if k > 0}

    return d


def quantile_n(x, n=10):
    """分位数权重

    Parameters
    ----------

    Returns
    -------

    """
    q = np.linspace(0, 1, n + 1)

    # 根据因子大小值进行百分位分层，数少就少分几层
    labels = qcut(x, q=q, rank=True, pct=True)

    # 因子分层，在这里就已经过滤只看几组，减少内存
    d = {k: np.where(labels == i, x, np.nan) for i, k in enumerate(q) if k > 0}

    return d



"""
公式转alpha191

国泰君安－基于短周期价量特征的多因子选股
"""

from ..imports import wide as W

CORR = W.ts_corr
REGSLOPE = W.LINEARREG_SLOPE
MEAN = W.SMA_TA
WMA = W.SMA_TA  # !!!WMA的公式没看懂，所以用另一个替代，以后再改
DECAYLINEAR = W.ts_decay_linear

CUMPROD = W.CUMPROD
FILTER = W.FILTER_191
RANK = W.rank
TSRANK = W.ts_rank
LessThan = W.less

IF = W.if_else
ABS = W.abs_
LOG = W.log  # 这里是用的自然对数
MAX = W.max_
MIN = W.min_
SIGN = W.sign

SMA = W.SMA_CN

COUNT = W.ts_count
DELTA = W.ts_delta
TSMAX = W.ts_max
HIGHDAY = W.ts_arg_max
TSMIN = W.ts_min
LOWDAY = W.ts_arg_min
MA = W.ts_mean
PROD = W.ts_product
DELAY = W.ts_delay
SUM = W.ts_sum
SUMIF = W.SUMIF  # 注意，SUMIF参数的位置常用的方式不同

REGBETA = W.SLOPE_YX
REGRESI = W.REGRESI4

COVIANCE = W.ts_covariance
STD = W.ts_std_dev  # 引入的是全体标准差



"""
Cross Sectional Operators
"""
import numpy as np

from .arithmetic import add, abs_
from .. import bn_wraps as bn
from ..regress import multiple_regress
from ..utils import pd_to_np


def normalize(x, useStd=False, limit=0.0):
    """Calculates the mean value of all valid alpha values for a certain date, then subtracts that mean from each element."""
    axis = x.ndim - 1
    t1 = np.nanmean(x, axis=axis, keepdims=True)
    if useStd:
        # 这里用ddof=1才能与文档示例的数值对应上
        t2 = np.nanstd(x, axis=axis, keepdims=True, ddof=1)
        r = (x - t1) / t2
    else:
        r = (x - t1)

    return r if limit == 0 else np.clip(r, -limit, limit)


def one_side(x, side='long'):
    """Shifts all instruments up or down so that the Alpha becomes long-only or short-only
(if side = short), respectively."""
    # TODO: 这里不确定，需再研究
    # [-1, 0, 1]+1=[0, 1, 2]
    # max([-1, 0, 1], 0)=[0,0,1]
    if side == 'long':
        return np.maximum(x, 0)
    else:
        return np.minimum(x, 0)


def quantile(x, driver='gaussian', sigma=1.0):
    """Rank the raw vector, shift the ranked Alpha vector, apply distribution ( gaussian, cauchy, uniform ). If driver is uniform, it simply subtract each Alpha value with the mean of all Alpha values in the Alpha vector."""
    pass


def rank(x, rate=2, pct=True):
    """Ranks the input among all the instruments and returns an equally distributed number between 0.0 and 1.0. For precise sort, use the rate as 0."""
    axis = x.ndim - 1
    t1 = bn.nanrankdata(x, axis=axis)

    if pct:
        t2 = np.nansum(~np.isnan(x), axis=axis, keepdims=True)

        return t1 / t2
    else:
        return t1


def rank_by_side(x, rate=2, scale=1):
    """Ranks positive and negative input separately and scale to book. For precise sorting use rate=0."""
    pass


def generalized_rank(open, m=1):
    """The idea is that difference between instrument values raised to the power of m is added to the rank of instrument with bigger value and subtracted from the rank of instrument with lesser value. More details in the notes at the end of page."""
    pass


def regression_neut(y, x):
    """Conducts the cross-sectional regression on the stocks with Y as target and X as the independent variable."""
    return multiple_regress(y, x, add_constant=True)[0]


def regression_proj(y, x):
    """Conducts the cross-sectional regression on the stocks with Y as target and X as the independent variable."""
    return multiple_regress(y, x, add_constant=True)[1]


def scale(x, scale=1, longscale=1, shortscale=1):
    """Scales input to booksize. We can also scale the long positions and short positions to separate scales by mentioning additional parameters to the operator."""
    axis = x.ndim - 1
    if longscale != 1 or shortscale != 1:
        L = np.where(x > 0, x, np.nan)
        S = np.where(x < 0, x, np.nan)

        sum_l = np.nansum(abs_(L), axis=axis, keepdims=True)
        sum_s = np.nansum(abs_(S), axis=axis, keepdims=True)

        with np.errstate(divide='ignore', invalid='ignore'):
            return add(L / sum_l * longscale, S / sum_s * shortscale, filter=True)
    else:
        sum_x = np.nansum(abs_(x), axis=axis, keepdims=True)
        with np.errstate(divide='ignore', invalid='ignore'):
            return x / sum_x * scale


def scale_down(x, constant=0):
    """Scales all values in each day proportionately between 0 and 1 such that minimum value maps to 0 and maximum value maps to 1. Constant is the offset by which final result is subtracted."""
    axis = x.ndim - 1
    m1 = np.nanmin(x, axis=axis, keepdims=True)
    m2 = np.nanmax(x, axis=axis, keepdims=True)

    return (x - m1) / (m2 - m1) - constant


def truncate(x, maxPercent=0.01):
    """Operator truncates all values of x to maxPercent. Here, maxPercent is in decimal notation."""
    axis = x.ndim - 1
    t1 = np.nansum(x, axis=axis, keepdims=True) * maxPercent

    return np.minimum(x, t1)


def vector_neut(x, y):
    """For given vectors x and y, it finds a new vector x* (output) such that x* is orthogonal to y."""
    pass


def vector_proj(x, y):
    """Returns vector projection of x onto y. Algebraic and geometric details can be found on wiki"""
    pass


def winsorize(x, std=4):
    """Winsorizes x to make sure that all values in x are between the lower and upper limits, which are specified as multiple of std. Details can be found on wiki"""
    x = pd_to_np(x, copy=False)
    axis = x.ndim - 1
    _mean = np.nanmean(x, axis=axis, keepdims=True)
    _std = np.nanstd(x, axis=axis, keepdims=True, ddof=0) * std

    return np.clip(x, _mean - _std, _mean + _std)


def zscore(x):
    """Z-score is a numerical measurement that describes a value's relationship to the mean of a group of values. Z-score is measured in terms of standard deviations from the mean"""
    x = pd_to_np(x, copy=False)
    axis = x.ndim - 1
    _mean = np.nanmean(x, axis=axis, keepdims=True)
    _std = np.nanstd(x, axis=axis, keepdims=True, ddof=0)

    return (x - _mean) / _std


def rank_gmean_amean_diff(*args):
    """Operator returns difference of geometric mean and arithmetic mean of cross sectional rank of inputs."""

    # TODO: 输入输出的形式还没搞清，核心功能已经实现先放这
    def _gmean(x):
        return np.exp(np.nanmean(np.log(x)))

    inputs = rank(np.array(args))
    inputs = np.array(args)
    return _gmean(inputs) - np.nanmean(inputs)



"""
Logical Operators
"""
from functools import reduce

import numpy as np


def and_(*args):
    """Logical AND operator, returns true if both operands are true and returns false otherwise"""
    return reduce(lambda x, y: np.logical_and(x, y), list(args))


def or_(*args):
    """Logical OR operator returns true if either or both inputs are true and returns false otherwise"""
    return reduce(lambda x, y: np.logical_or(x, y), args)


def equal(input1, input2):
    """Returns true if both inputs are same and returns false otherwise"""
    return input1 == input2


def negate(input):
    """The result is true if the converted operand is false; the result is false if the converted operand is true"""
    return ~input


def less(input1, input2):
    """If input1 < input2 return true, else return false"""
    # ValueError: Can only compare identically-labeled Series objects
    return input1 - input2 < 0


def if_else(input1, input2, input3):
    """If input1 is true then return input2 else return input3."""
    return np.where(input1, input2, input3)


def is_not_nan(input):
    """If (input != NaN) return 1 else return 0"""
    return input == input


def is_nan(input):
    """If (input == NaN) return 1 else return 0"""
    return input != input


def is_finite(input):
    """If (input NaN or input == INF) return 0, else return 1"""
    return np.isinf(input)


def is_not_finite(input):
    """If (input NAN or input == INF) return 1 else return 0"""
    return ~np.isinf(input)



"""
公式转alpha191

国泰君安－基于短周期价量特征的多因子选股
"""

from ..imports import long as L

CORR = L.ts_corr
REGSLOPE = L.LINEARREG_SLOPE
MEAN = L.SMA_TA
WMA = L.SMA_TA  # !!!WMA的公式没看懂，所以用另一个替代，以后再改
DECAYLINEAR = L.ts_decay_linear

CUMPROD = L.CUMPROD
FILTER = L.FILTER_191
RANK = L.rank
TSRANK = L.ts_rank
LessThan = L.less

IF = L.if_else
ABS = L.abs_
LOG = L.log  # 这里是用的自然对数
MAX = L.max_
MIN = L.min_
SIGN = L.sign

SMA = L.SMA_CN

COUNT = L.ts_count
DELTA = L.ts_delta
TSMAX = L.ts_max
HIGHDAY = L.ts_arg_max
TSMIN = L.ts_min
LOWDAY = L.ts_arg_min
MA = L.ts_mean
PROD = L.ts_product
DELAY = L.ts_delay
SUM = L.ts_sum
SUMIF = L.SUMIF  # 注意，SUMIF参数的位置常用的方式不同

REGBETA = L.SLOPE_YX
REGRESI = L.REGRESI4

COVIANCE = L.ts_covariance
STD = L.ts_std_dev  # 引入的是全体标准差



import numba
import numpy as np

from . import SUM
from .. import numba_cache
from ..nb import numpy_rolling_apply, _rolling_func_1_nb
from ..utils import np_to_pd, num_to_np, pd_to_np


def CROSS(S1, S2):
    """向上金叉"""
    # 可能输入单个数字，需预处理
    S1 = num_to_np(S1, S2)
    S2 = num_to_np(S2, S1)

    arr = np.zeros_like(S1, dtype=bool)
    arr[1:] = (S1 <= S2)[:-1] & (S1 > S2)[1:]
    return arr


def EVERY(real, timeperiod):
    """最近timeperiod天是否全为True

    EVERY(real, timeperiod=5)
    """
    return SUM(real, timeperiod) == timeperiod


def EXIST(real, timeperiod):
    """最近timeperiod天是否存在一天为True

    EXIST(real, timeperiod=5)
    """
    return SUM(real, timeperiod) > 0


def BETWEEN(S, A, B):
    """BETWEEN(A,B,C)表示A处于B和C之间时返回1,否则返回0"""
    return ((A < S) & (S < B)) | ((A > S) & (S > B))


def VALUEWHEN(S, X):
    """条件跟随函数。当COND条件成立时,取X的当前值,否则取VALUEWHEN的上个值"""
    return np_to_pd(np.where(S, X, np.nan)).ffill()


def LAST(real, n, m):
    """LAST(X,A,B)，A>B，表示从前A日到前B日一致满足X条件

    LAST(real, n=20, m=10)
    """

    @numba.jit(nopython=True, cache=numba_cache, nogil=True)
    def _last_nb(arr, n, m):
        """LAST(X,A,B)，A>B，表示从前A日到前B日一致满足X条件"""
        return np.all(arr[:n - m])

    return numpy_rolling_apply([pd_to_np(real)], n, _rolling_func_1_nb, _last_nb, n, m)



"""
对指标的算子化包装
1. 包装成只支持 宽表 输入，输出是特殊格式，需要处理得到输出
2. 简化参数输入，命名参数也可当成位置参数输入
3. 通过堆叠的方法，自动跳过停牌

!!!函数太多，又想要智能提示，只能手工按需补充
"""
from .long import group_neutralize
from .. import talib as ta
from ..alphas.alpha import CUMPROD
from ..alphas.alpha import FILTER_191
from ..ema import SMA_CN
from ..regress import REGRESI
from ..regress import SLOPE_YX
from ..tdx.reference import FILTER as FILTER_TDX
from ..tdx.reference import SUMIF
from ..utils_wide import wide_wraps, long_wraps
from ..wq.arithmetic import abs_
from ..wq.arithmetic import log
from ..wq.arithmetic import max_
from ..wq.arithmetic import min_
from ..wq.arithmetic import sign
from ..wq.arithmetic import signed_power
from ..wq.cross_sectional import rank
from ..wq.cross_sectional import scale
from ..wq.logical import if_else
from ..wq.logical import less
from ..wq.time_series import ts_arg_max
from ..wq.time_series import ts_arg_min
from ..wq.time_series import ts_corr
from ..wq.time_series import ts_count
from ..wq.time_series import ts_covariance
from ..wq.time_series import ts_decay_linear
from ..wq.time_series import ts_delay
from ..wq.time_series import ts_delta
from ..wq.time_series import ts_max
from ..wq.time_series import ts_mean
from ..wq.time_series import ts_min
from ..wq.time_series import ts_product
from ..wq.time_series import ts_rank
from ..wq.time_series import ts_std_dev
from ..wq.time_series import ts_sum

# 二维TALIB
_ta2d = ta.init(mode=2, skipna=False, to_globals=False)

# TALIB, 多输入
CORREL = wide_wraps(_ta2d.CORREL, input_num=2, to_kwargs={2: 'timeperiod'})
ATR = wide_wraps(_ta2d.ATR, input_num=3, to_kwargs={3: 'timeperiod'})

# TALIB, 单输入
LINEARREG_SLOPE = wide_wraps(_ta2d.LINEARREG_SLOPE)
SMA_TA = wide_wraps(_ta2d.SMA)
WMA = wide_wraps(_ta2d.WMA)

# WorldQuant，时序
ts_arg_max = wide_wraps(ts_arg_max, to_kwargs={1: 'd'})
ts_arg_min = wide_wraps(ts_arg_min, to_kwargs={1: 'd'})
ts_corr = wide_wraps(ts_corr, input_num=2, to_kwargs={2: 'd'})
ts_count = wide_wraps(ts_count, to_kwargs={1: 'd'})
ts_covariance = wide_wraps(ts_covariance, input_num=2, to_kwargs={2: 'd'})
ts_decay_linear = wide_wraps(ts_decay_linear, to_kwargs={1: 'd'})
ts_delay = wide_wraps(ts_delay, to_kwargs={1: 'd'})
ts_delta = wide_wraps(ts_delta, to_kwargs={1: 'd'})
ts_max = wide_wraps(ts_max, to_kwargs={1: 'd'})
ts_mean = wide_wraps(ts_mean, to_kwargs={1: 'd'})
ts_min = wide_wraps(ts_min, to_kwargs={1: 'd'})
ts_product = wide_wraps(ts_product, to_kwargs={1: 'd'})
ts_rank = wide_wraps(ts_rank, to_kwargs={1: 'd'})
ts_std_dev = wide_wraps(ts_std_dev, to_kwargs={1: 'd'})
ts_sum = wide_wraps(ts_sum, to_kwargs={1: 'd'})

# WorldQuant，横截面
rank = wide_wraps(rank, direction=None, to_kwargs={})
scale = wide_wraps(scale, direction=None, to_kwargs={1: 'scale'})

# WorldQuant
less = wide_wraps(less, input_num=2, to_kwargs={})
signed_power = wide_wraps(signed_power, direction=None, input_num=2, to_kwargs={})
if_else = wide_wraps(if_else, direction=None, input_num=3, to_kwargs={})
abs_ = wide_wraps(abs_, direction=None, to_kwargs={})
log = wide_wraps(log, direction=None, to_kwargs={})
max_ = wide_wraps(max_, direction=None, input_num=2, to_kwargs={})
min_ = wide_wraps(min_, direction=None, input_num=2, to_kwargs={})
sign = wide_wraps(sign, direction=None, to_kwargs={})

#
CUMPROD = wide_wraps(CUMPROD, to_kwargs={})
FILTER_191 = wide_wraps(FILTER_191, input_num=2, to_kwargs={})
FILTER_TDX = wide_wraps(FILTER_TDX, input_num=2, to_kwargs={1: 'N'})

#
SMA_CN = wide_wraps(SMA_CN, to_kwargs={1: 'timeperiod', 2: 'M'})

# 通达信导入
SUMIF = wide_wraps(SUMIF, input_num=2, to_kwargs={2: 'timeperiod'})

SLOPE_YX = wide_wraps(SLOPE_YX, input_num=2, to_kwargs={2: 'timeperiod'})
REGRESI4 = wide_wraps(REGRESI, input_num=4, to_kwargs={4: 'timeperiod'})

# 长表转宽表
# 行业中性化
group_neutralize = long_wraps(group_neutralize, direction='right')



"""

"""
import numpy as np


def FILTER_191(A, condition):
    """注意：与TDX中的FILTER不同"""
    return np.where(condition, A, 0)


def CUMPROD(A):
    return np.cumprod(A, axis=0)


def CUMSUM(A):
    return np.cumsum(A, axis=0)



import numpy as np

from .utils import pd_to_np
from .wq.cross_sectional import winsorize, zscore, scale


def winsorize_mad(x, n=3, constant=1.4826):
    """缩尾去极值，MAD法

    Parameters
    ----------
    x: array
        需要缩尾去极值的数据
    n: float
        设置范围大小
    constant: float
        比例因子。为了能将MAD当作标准差估计的一种一致估计量，使用两者之间的关联。其中比例因子常量k的值取决于分布类型，在正态分布下该常量约等于1.4826

        .. math:: \hat {\sigma } = k * MAD

    Returns
    -------
    array
        `x` 经过MAD法缩尾去极值处理后的新数据

    References
    ----------
    https://en.wikipedia.org/wiki/Median_absolute_deviation


    """
    x = pd_to_np(x, copy=False)
    axis = x.ndim - 1
    _median = np.nanmedian(x, axis=axis, keepdims=True)
    _mad = np.nanmedian(abs(x - _median), axis=axis, keepdims=True)
    _mad = _mad * constant * n

    return np.clip(x, _median - _mad, _median + _mad)


def winsorize_quantile(x, min_=0.1, max_=0.9):
    """缩尾去极值，分位数法

    Parameters
    ----------
    x: array
        需要缩尾去极值的数据
    min_: float
        设置下界分位数
    max_: float
        设置上界分位数

    Returns
    -------
    array
        `x` 经过分位数法缩尾去极值处理后的新数据

    """
    x = pd_to_np(x, copy=False)
    axis = x.ndim - 1
    q = np.nanquantile(x, [min_, max_], axis=axis, keepdims=True)
    return np.clip(x, q[0], q[1])


"""
删除极值
"""


def drop_quantile(x, min_=0.1, max_=0.9):
    """删除去极值，分位数法

    Parameters
    ----------
    x: array
        需要删除去极值的数据
    min_: float
        设置下界分位数
    max_: float
        设置上界分位数

    Returns
    -------
    array
        `x` 经过分位数法删除去极值处理后的新数据

    """
    x = pd_to_np(x, copy=False)
    axis = x.ndim - 1

    q = np.nanquantile(x, [min_, max_], axis=axis, keepdims=True)
    x = np.where((x < q[0]) | (x > q[1]), np.nan, x)
    return x


"""
标准化
"""


def fill_na(x):
    """用中位数填充，还是用平均值填充？

    -1到1归一化的值，用0填充也行吧？
    """
    x = x.copy()
    x[np.isnan(x)] = np.nanmedian(x)
    return x


# worldquant中函数的别名
winsorize_3sigma = winsorize
standardize_zscore = zscore
standardize_minmax = scale



"""
优先实现WorldQuant BRAIN中Fast Expression的函数
然后再实现通达信的函数
"""

from ..wq.arithmetic import abs_ as ABS
from ..wq.arithmetic import add as ADD
from ..wq.arithmetic import divide as DIV
from ..wq.arithmetic import log as LN  # 自然对数
from ..wq.arithmetic import log10 as LOG  # 10为底的对数
from ..wq.arithmetic import max_ as MAX
from ..wq.arithmetic import mean as MEAN
from ..wq.arithmetic import min_ as MIN
from ..wq.arithmetic import multiply as MUL
from ..wq.arithmetic import round_ as ROUND
from ..wq.arithmetic import sign as SGN
from ..wq.arithmetic import subtract as SUB
from ..wq.logical import if_else as IF
from ..wq.time_series import ts_count as COUNT
from ..wq.time_series import ts_delay as REF
from ..wq.time_series import ts_delta as DIFF
from ..wq.time_series import ts_max as HHV
from ..wq.time_series import ts_mean as MA
from ..wq.time_series import ts_min as LLV
from ..wq.time_series import ts_sum as SUM
from ..wq.cross_sectional import rank as RANK

ABS
MAX
MIN
REF
HHV
MA
LLV
SUM
ADD
SUB
MUL
DIV
ROUND
MEAN
LN
LOG
SGN
DIFF
IF
COUNT
RANK



"""
Group Operators

分组计算
返回的组内是一个值，还是多个值？
1. group_count等等肯定是一个值。输出索引是date,group
2. group_rankt等等是输出多个值。输出索引是date,asset

"""
import numpy as np

from .. import BY_GROUP
from ..utils_long import dataframe_groupby_apply


def group_backfill(x, group, d, std=4.0):
    """If a certain value for a certain date and instrument is NaN, from the set of same group instruments, calculate winsorized mean of all non-NaN values over last d days."""
    pass


def group_count(x, group):
    """Gives the number of instruments in the same group (e.g. sector) which have valid values of x. For example, x=1 gives the number of instruments in each group (without regard for whether any particular field has valid data)."""
    func = dataframe_groupby_apply(len, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_extra(x, weight, group):
    """Replaces NaN values by their corresponding group means."""
    pass
    # func = dataframe_groupby_apply(None, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={}, dropna=False)
    #
    # return func(x, group)


def group_max(x, group):
    """Maximum of x for all instruments in the same group."""
    func = dataframe_groupby_apply(np.max, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_mean(x, weight, group):
    """All elements in group equals to the mean value of the group. Mean = sum(data*weight) / sum(weight) in each group."""

    # TODO: 这里的权重是与x等长的权重序列，还是与股票数一样的权重？
    def _mean(x, weight):
        return np.average(x, weights=weight)

    func = dataframe_groupby_apply(_mean, by=BY_GROUP, to_df=[0, 1, 'group'], to_kwargs={})

    return func(x, weight, group)


def group_median(x, group):
    """All elements in group equals to the median value of the group."""
    func = dataframe_groupby_apply(np.median, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_min(x, group):
    """All elements in group equals to the min value of the group."""
    func = dataframe_groupby_apply(np.min, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_neutralize(x, group):
    """Neutralizes Alpha against groups. These groups can be subindustry, industry, sector, country or a constant."""

    def _demean(x):
        """行业中性化，需要与groupby配合使用

        RuntimeWarning: Mean of empty slice
        nanmean在全nan时报此警告。这个警告还不好屏蔽
        """
        return x - np.nanmean(x)

    func = dataframe_groupby_apply(_demean, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_normalize(x, group, constantCheck=False, tolerance=0.01, scale=1):
    """Normalizes input such that each group's absolute sum is 1."""

    # 发现这里的group_normalize其实像scale，也就是不少地方混乱
    def _normalize(x, scale):
        sum_x = np.sum(abs(x))
        return x / sum_x * scale

    func = dataframe_groupby_apply(_normalize, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={2: 'scale'})

    return func(x, group, scale)


def group_percentage(x, group, percentage=0.5):
    """All elements in group equals to the value over the percentage of the group.
Percentage = 0.5 means value is equal to group_median(x, group)"""

    func = dataframe_groupby_apply(np.quantile, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={2: 'q'})

    return func(x, group, percentage)


def group_vector_proj(x, y, g):
    """
Similar to vector_proj(x, y) but x projection to y for each group which can be any classifier such as subindustry, industry, sector, etc. Refer wiki  for more details"""
    pass


def group_rank(x, group):
    """Each elements in a group is assigned the corresponding rank in this group"""
    func = dataframe_groupby_apply(lambda x: x.rank(pct=True), by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_scale(x, group):
    """Normalizes the values in a group to be between 0 and 1. (x - groupmin) / (groupmax - groupmin)"""

    def _scale(x):
        t1 = np.min(x)
        t2 = np.max(x)
        return (x - t1) / (t2 - t1)

    func = dataframe_groupby_apply(_scale, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_std_dev(x, group):
    """All elements in group equals to the standard deviation of the group."""

    func = dataframe_groupby_apply(np.std, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_sum(x, group):
    """Sum of x for all instruments in the same group."""
    func = dataframe_groupby_apply(np.sum, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)


def group_vector_neut(x, y, g):
    """Similar to vector_neut(x, y) but x neutralize to y for each group g which can be any classifier such as subindustry, industry, sector, etc."""
    pass


def group_zscore(x, group):
    """Calculates group Z-score - numerical measurement that describes a value's relationship to the mean of a group of values. Z-score is measured in terms of standard deviations from the mean. zscore = (data - mean) / stddev of x for each instrument within its group."""

    def _zscore(x):
        return (x - np.mean(x)) / np.std(x)

    func = dataframe_groupby_apply(_zscore, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})

    return func(x, group)



"""
本项目参考了WorldQuant BRAIN中Fast Expression的语法

References
----------
1. https://platform.worldquantbrain.com/learn/data-and-operators/operators
2. https://platform.worldquantbrain.com/learn/data-and-operators/detailed-operator-descriptions
"""
